/*
 *  Copyright (c) 2007,
 *  Commissariat a l'Energie Atomique (CEA)
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without modification,
 *  are permitted provided that the following conditions are met:
 *
 *   - Redistributions of source code must retain the above copyright notice, this
 *     list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright notice,
 *     this list of conditions and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *
 *   - Neither the name of CEA nor the names of its contributors may be used to
 *     endorse or promote products derived from this software without specific prior
 *     written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.
 *  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 *  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 *  OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Authors: Gilles Mouchard (gilles.mouchard@cea.fr)
 */
 
#ifndef __UNISIM_SERVICE_DEBUG_INLINE_DEBUGGER_INLINE_DEBUGGER_HH__
#define __UNISIM_SERVICE_DEBUG_INLINE_DEBUGGER_INLINE_DEBUGGER_HH__

#include <unisim/service/interfaces/memory_access_reporting.hh>
#include <unisim/service/interfaces/debug_control.hh>
#include <unisim/service/interfaces/disassembly.hh>
#include <unisim/service/interfaces/symbol_table_lookup.hh>
#include <unisim/service/interfaces/registers.hh>
#include <unisim/service/interfaces/memory.hh>
#include <unisim/service/interfaces/trap_reporting.hh>
#include <unisim/service/interfaces/loader.hh>
#include <unisim/service/interfaces/stmt_lookup.hh>
#include <unisim/service/interfaces/backtrace.hh>

#include <unisim/util/debug/breakpoint_registry.hh>
#include <unisim/util/debug/watchpoint_registry.hh>
#include <unisim/util/debug/profile.hh>
#include <unisim/util/loader/elf_loader/elf32_loader.hh>
#include <unisim/util/loader/elf_loader/elf64_loader.hh>

#include <unisim/kernel/service/service.hh>
#include <unisim/kernel/logger/logger.hh>

#include <inttypes.h>
#include <string>
#include <vector>
#include <queue>
#include <list>
#include <stack>

#ifdef WIN32
#include <windows.h>
#endif

namespace unisim {
namespace service {
namespace debug {
namespace inline_debugger {

using unisim::service::interfaces::DebugControl;
using unisim::service::interfaces::Disassembly;
using unisim::service::interfaces::MemoryAccessReporting;
using unisim::service::interfaces::MemoryAccessReportingControl;
using unisim::service::interfaces::SymbolTableLookup;
using unisim::service::interfaces::Registers;
using unisim::service::interfaces::Memory;
using unisim::service::interfaces::TrapReporting;
using unisim::service::interfaces::Loader;
using unisim::service::interfaces::StatementLookup;
using unisim::service::interfaces::BackTrace;

using unisim::util::debug::BreakpointRegistry;
using unisim::util::debug::Breakpoint;
using unisim::util::debug::WatchpointRegistry;
using unisim::util::debug::Watchpoint;
using unisim::util::debug::Symbol;
using unisim::util::debug::Statement;

using unisim::kernel::service::Service;
using unisim::kernel::service::ServiceExport;
using unisim::kernel::service::ServiceImport;
using unisim::kernel::service::Object;
using unisim::kernel::service::Client;
using unisim::kernel::service::Parameter;

using std::string;

typedef enum
{
	INLINE_DEBUGGER_MODE_WAITING_USER,
	INLINE_DEBUGGER_MODE_STEP,
	INLINE_DEBUGGER_MODE_CONTINUE,
	INLINE_DEBUGGER_MODE_CONTINUE_UNTIL,
	INLINE_DEBUGGER_MODE_RESET
} InlineDebuggerRunningMode;

class InlineDebuggerBase
{
public:
	InlineDebuggerBase();
	virtual ~InlineDebuggerBase();
protected:
	static bool trap;
private:
#ifndef WIN32
	static void (*prev_sig_int_handler)(int);
#endif
	static int alive_instances;
#ifdef WIN32
	static BOOL WINAPI ConsoleCtrlHandler(DWORD dwCtrlType);
#else
	static void SigIntHandler(int signum);
#endif
};

template <class ADDRESS>
class InlineDebugger :
	public Service<DebugControl<ADDRESS> >,
	public Service<MemoryAccessReporting<ADDRESS> >,
	public Service<TrapReporting>,
	public Client<MemoryAccessReportingControl>,
	public Client<Disassembly<ADDRESS> >,
	public Client<Memory<ADDRESS> >,
	public Client<Registers>,
	public Client<SymbolTableLookup<ADDRESS> >,
	public Client<Loader>,
	public Client<StatementLookup<ADDRESS> >,
	public Client<BackTrace<ADDRESS> >,
	public InlineDebuggerBase
{
public:
	ServiceExport<DebugControl<ADDRESS> > debug_control_export;
	ServiceExport<MemoryAccessReporting<ADDRESS> > memory_access_reporting_export;
	ServiceExport<TrapReporting> trap_reporting_export;
	ServiceImport<Disassembly<ADDRESS> > disasm_import;
	ServiceImport<Memory<ADDRESS> > memory_import;
	ServiceImport<MemoryAccessReportingControl> memory_access_reporting_control_import;
	ServiceImport<Registers> registers_import;
	ServiceImport<SymbolTableLookup<ADDRESS> > **symbol_table_lookup_import;
	ServiceImport<Loader> **loader_import;
	ServiceImport<StatementLookup<ADDRESS> > **stmt_lookup_import;
	ServiceImport<BackTrace<ADDRESS> > **backtrace_import;
	
	InlineDebugger(const char *name, Object *parent = 0);
	virtual ~InlineDebugger();

	// MemoryAccessReportingInterface
	virtual void ReportMemoryAccess(typename MemoryAccessReporting<ADDRESS>::MemoryAccessType mat, typename MemoryAccessReporting<ADDRESS>::MemoryType mt, ADDRESS addr, uint32_t size);
	virtual void ReportFinishedInstruction(ADDRESS next_addr);
	virtual void ReportTrap();
	virtual void ReportTrap(const unisim::kernel::service::Object &obj);
	virtual void ReportTrap(const unisim::kernel::service::Object &obj,
							const std::string &str);
	virtual void ReportTrap(const unisim::kernel::service::Object &obj,
							const char *c_str);
	
	// DebugControlInterface
	virtual typename DebugControl<ADDRESS>::DebugCommand FetchDebugCommand(ADDRESS cia);

	virtual bool EndSetup();
	virtual void OnDisconnect();
private:
	unisim::kernel::logger::Logger logger;
	unsigned int memory_atom_size;
	unsigned int num_loaders;
	std::string search_path;
	std::string init_macro;
	std::string output;
	Parameter<unsigned int> param_memory_atom_size;
	Parameter<unsigned int> param_num_loaders;
	Parameter<std::string> param_search_path;
	Parameter<std::string> param_init_macro;
	Parameter<std::string> param_output;

	BreakpointRegistry<ADDRESS> breakpoint_registry;
	WatchpointRegistry<ADDRESS> watchpoint_registry;
	unisim::util::debug::Profile<ADDRESS> program_profile;
	unisim::util::debug::Profile<ADDRESS> data_read_profile;
	unisim::util::debug::Profile<ADDRESS> data_write_profile;
	InlineDebuggerRunningMode running_mode;
	std::vector<unisim::util::loader::elf_loader::Elf32Loader<ADDRESS> *> elf32_loaders;
	std::vector<unisim::util::loader::elf_loader::Elf64Loader<ADDRESS> *> elf64_loaders;

	ADDRESS disasm_addr;
	ADDRESS dump_addr;
	ADDRESS cont_until_addr;

	std::list<std::string> exec_queue;
	string prompt;
	char *hex_addr_fmt;
	char *int_addr_fmt;
	std::string last_line;
	std::string line;
	std::ostream *output_stream;
	std::ostream *std_output_stream;
	std::ostream *std_error_stream;

	void Tokenize(const std::string& str, std::vector<std::string>& tokens);
	bool ParseAddr(const char *s, ADDRESS& addr);
	bool ParseAddrRange(const char *s, ADDRESS& addr, unsigned int& size);
	bool GetLine(const char *prompt, std::string& line, bool& interactive);
	bool IsBlankLine(const std::string& line);
	bool IsQuitCommand(const char *cmd);
	bool IsStepCommand(const char *cmd);
	bool IsNextCommand(const char *cmd);
	bool IsContinueCommand(const char *cmd);
	bool IsDisasmCommand(const char *cmd);
	bool IsBreakCommand(const char *cmd);
	bool IsWatchCommand(const char *cmd);
	bool IsDeleteCommand(const char *cmd);
	bool IsDeleteWatchCommand(const char *cmd);
	bool IsDumpCommand(const char *cmd);
	bool IsEditCommand(const char *cmd);
	bool IsHelpCommand(const char *cmd);
	bool IsResetCommand(const char *cmd);
	bool IsMonitorCommand(const char *cmd, const char *format = 0);
	bool IsRegisterCommand(const char *cmd, const char *format = 0);
	bool IsStatisticCommand(const char *cmd, const char *format = 0);
	bool IsParameterCommand(const char *cmd, const char *format = 0);
	bool IsSymbolCommand(const char *cmd);
	bool IsMonitorSetCommand(const char *cmd);
	bool IsProfileCommand(const char *cmd);
	bool IsLoadCommand(const char *cmd);
	bool IsBackTraceCommand(const char *cmd);
	bool IsLoadSymbolTableCommand(const char *cmd);
	bool IsListSymbolsCommand(const char *cmd);
	bool IsMacroCommand(const char *cmd);

	void Help();
	void Disasm(ADDRESS addr, int count);
	bool HasBreakpoint(ADDRESS addr);
	void SetBreakpoint(ADDRESS addr);
	void SetReadWatchpoint(ADDRESS addr, uint32_t size);
	void SetWriteWatchpoint(ADDRESS addr, uint32_t size);
	void DeleteBreakpoint(ADDRESS addr);
	void DeleteReadWatchpoint(ADDRESS addr, uint32_t size);
	void DeleteWriteWatchpoint(ADDRESS addr, uint32_t size);
	void DumpBreakpoints();
	void DumpWatchpoints();
	void DumpMemory(ADDRESS addr);
	bool EditMemory(ADDRESS addr);
	void DumpSymbols(const typename std::list<const unisim::util::debug::Symbol<ADDRESS> *>& symbols, const char *name = 0);
	void DumpSymbols(const char *name = 0);
	void MonitorGetFormat(const char *cmd, char &format);
	void DumpVariables(const char *cmd, const char *name = 0, typename unisim::kernel::service::VariableBase::Type type = unisim::kernel::service::VariableBase::VAR_VOID);
	void DumpVariable(const char *cmd, const unisim::kernel::service::VariableBase *variable);
	void SetVariable(const char *name, const char *value);
	void DumpProgramProfile();
	void DumpDataProfile(bool write);
	void DumpAvailableLoaders();
	void Load(const char *loader_name);
	std::string SearchFile(const char *filename);
	void LoadSymbolTable(const char *filename);
	void LoadMacro(const char *filename);
	void DumpSource(const char *filename, unsigned int lineno, unsigned int colno, unsigned int count);
	void DumpBackTrace(ADDRESS cia);
	const Symbol<ADDRESS> *FindSymbolByAddr(ADDRESS addr);
	const Symbol<ADDRESS> *FindSymbolByName(const char *s);
	const Statement<ADDRESS> *FindStatement(ADDRESS addr);
	const Statement<ADDRESS> *FindStatement(const char *filename, unsigned int lineno, unsigned int colno);
};

} // end of namespace inline_debugger
} // end of namespace debug
} // end of namespace service
} // end of namespace unisim

#endif
